from arkitekt.actors.functional import (
    FunctionalFuncActor,
    FunctionalGenActor,
    FunctionalThreadedFuncActor,
    FunctionalThreadedGenActor,
)
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from koil import koil
from mikroj.actors.define_macro import MacroDefinition


class MacroActor:
    async def run_macro(self, **kwargs):
        return self.helper.py.run_macro(self.macro.code, **kwargs)


class FuncMacroActor(FunctionalThreadedFuncActor):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)



import imagej
import scyjava
import scyjava
import xarray as xr
import dask
from mikro import Representation
import os


class BaseImageJHelper:
    def __init__(
        self, headless=False, version="2.1.0", plugins=[], plugins_dir="plugins"
    ) -> None:
        # concatenade version plus plugins
        # build = [version] + plugins if len(plugins) > 0 else version
        # if plugins_dir:
        #   path = os.path.join(os.getcwd(),plugins_dir)
        #    scyjava.config.add_option(f'-Dplugins.dir={path}')

        self.headless = headless
        print(f"Initializing with version {version}")
        plugins_dir = "/home/jhnnsrs/Fiji.app/plugins"
        scyjava.config.add_option(f"-Dplugins.dir={plugins_dir}")

        self._ij = imagej.init(version, headless=headless)
        if not headless:
            self._ij.ui().showUI()
        super().__init__()

    @property
    def py(self):
        return self._ij.py

    @property
    def ui(self):
        return self._ij.ui()

    @property
    def ij(self):
        return self._ij

    def show_xarray(self, xarray: xr.DataArray):
        if dask.is_dask_collection(xarray.data):
            jimage = self.py.to_java(xarray.compute())
        else:
            jimage = self.py.to_java(xarray)

        self.ui.show(xarray.name, jimage)

    def displayRep(self, rep: Representation):
        image = rep.data.squeeze()

        if dask.is_dask_collection(image.data):
            jimage = self.py.to_java(image.compute())
        else:
            jimage = self.py.to_java(image)

        # Convert the Image to Image
        self.ui.show(rep.name, jimage)


RUNNING_HELPER = None


def get_running_helper() -> BaseImageJHelper:
    global RUNNING_HELPER
    if RUNNING_HELPER is None:
        print("Happening")
        RUNNING_HELPER = BaseImageJHelper()
    return RUNNING_HELPER


def set_running_helper(instance: BaseImageJHelper):
    global RUNNING_HELPER
    RUNNING_HELPER = instance



from typing import Optional, Union

import pydantic
from arkitekt.schema.node import Node
from mikroj.parsers.base import Parser
from mikroj.parsers.definition.base import DefinitionParser
from mikroj.parsers.code.base import CodeParser
import pathlib
from functools import reduce
from mikroj.parsers.meta import MDMeta, parse_md_meta
from mikroj.registries.actor import get_current_macro_actor_registry
from pydantic import BaseModel
import logging
import os


logger = logging.getLogger(__name__)


class QueryNodeDefinition(BaseModel):
    package: Optional[str]
    interface: Optional[str]
    q: Optional[str]


class MacroRegistry:
    def __init__(self) -> None:
        self.registered_macros = {}
        self.registered_definition_parser = {}
        self.registered_code_parser = {}

    def register_definition_parser(self, className, parser):
        self.registered_definition_parser[className] = parser

    def register_code_parser(self, className, parser):
        self.registered_code_parser[className] = parser

    def scan_folder(self, folder_path="macros"):
        assert os.path.exists(folder_path), f"Folder {folder_path} does not exist"

        pathlist = pathlib.Path(folder_path).rglob("*.md")
        macro_list = []
        for path in pathlist:
            # because path is object not string
            path_in_str = str(path)
            logger.debug(f"Found file {path}")
            with open(path_in_str) as f:
                try:
                    stream = f.read()

                    meta = parse_md_meta(stream)

                    definition_parsers = [
                        self.registered_definition_parser[parser]
                        for parser in meta.get_definition_parsers()
                    ]
                    code_parsers = [
                        self.registered_code_parser[parser]
                        for parser in meta.get_code_parsers()
                    ]

                    definitions = [
                        parser(stream=stream).get_definition()
                        for parser in definition_parsers
                    ]
                    codes = [
                        parser(stream=stream).get_code() for parser in code_parsers
                    ]

                    combined_definition = reduce(
                        lambda x, y: {**x, **y} if x else y, definitions, {}
                    )

                    if (
                        "template" in combined_definition
                        and combined_definition["template"] == "yes"
                    ):
                        definition = QueryNodeDefinition(**combined_definition)
                    else:
                        definition = Node(**combined_definition)

                    actorBuilder = lambda: get_current_macro_actor_registry().get_actor(
                        meta.actor
                    )(macro=codes[0])
                    macro_list.append((definition, actorBuilder))

                except pydantic.error_wrappers.ValidationError as e:
                    logger.exception(e)
                    logger.error(f"Omiting file {path}")
                    continue

        return macro_list


def register_definition_parser(className: str):
    def rea_decorator(parser):
        assert issubclass(parser, Parser), "Parser must subclass Parser"
        logger.info(f"Registering Definition Parser {parser} for {className}")
        get_current_macro_registry().register_definition_parser(className, parser)
        return parser

    return rea_decorator


def register_code_parser(className: str):
    def rea_decorator(parser):
        assert issubclass(parser, Parser), "Parser must subclass Parser"
        logger.info(f"Registering Code parser {parser} for {className}")
        get_current_macro_registry().register_code_parser(className, parser)
        return parser

    return rea_decorator


MACRO_REGISTRY = None


def get_current_macro_registry(register_defaults=True) -> MacroRegistry:
    global MACRO_REGISTRY
    if MACRO_REGISTRY is None:
        MACRO_REGISTRY = MacroRegistry()

    return MACRO_REGISTRY



from arkitekt.packers.structure import BoundType
from arkitekt.schema.ports import (
    IntArgPort,
    IntReturnPort,
    StringArgPort,
    StringKwargPort,
    StructureArgPort,
    IntArgPort,
    IntReturnPort,
    StructureReturnPort,
)
from mikro.schema import Representation


class PortMatcher:
    def __init__(self) -> None:
        self.typeMap = {}

    def register(self, type, imagejType, argBuilder):
        self.typeMap.setdefault(type, {})[imagejType] = argBuilder

    def argbuilder_for_type(self, imagejType):
        return self.typeMap["arg"][imagejType]

    def kwargbuilder_for_type(self, imagejType):
        return self.typeMap["kwarg"][imagejType]

    def returnbuilder_for_type(self, imagejType):
        return self.typeMap["return"][imagejType]


def string_kwarg_builder(
    default=None,
    **kwargs,
):
    return StringKwargPort(**kwargs, defaultString=default.replace('"', ""))


def imageplus_arg_builder(
    **kwargs,
):
    return StructureArgPort(
        **kwargs,
        identifier="representation",
        widget=Representation.get_structure_meta().widget,
    )


def imageplus_return_builder(
    **kwargs,
):
    return StructureReturnPort(
        **kwargs,
        identifier="representation",
    )


MATCHER = PortMatcher()
MATCHER.register("arg", "String", StringArgPort)
MATCHER.register("kwarg", "String", string_kwarg_builder)
MATCHER.register("return", "Integer", IntReturnPort)
MATCHER.register("arg", "ImagePlus", imageplus_arg_builder)
MATCHER.register("return", "ImagePlus", imageplus_return_builder)
MATCHER.register("arg", "Dataset", imageplus_arg_builder)
MATCHER.register("return", "Dataset", imageplus_return_builder)


def get_current_matcher():
    global MATCHER
    return MATCHER



from mikro.schema import Representation
from mikroj.registries.helper import BaseImageJHelper, set_running_helper
import dask
import xarray as xr


class ImageJHelper(BaseImageJHelper):
    def __init__(
        self, headless=False, bind=True, version="sc.fiji:fiji", plugins=[]
    ) -> None:
        if bind:
            set_running_helper(self)

        super().__init__(
            headless=headless, version="/home/jhnnsrs/Fiji.app", plugins=plugins
        )



from mikroj.parsers.base import CodeBlockParser
from mikroj.parsers.definition.base import DefinitionParser
from mikroj.registries.macro import register_definition_parser
import yaml

@register_definition_parser("python")
class PythonDefinition(DefinitionParser):
    language = "language-python"

    def __init__(self, *args, stream=None, **kwargs) -> None:
        super().__init__(*args, stream=stream, **kwargs)
        self.parser = CodeBlockParser(stream, self.language)

    def get_definition(self):
        






        return yaml.load(self.parser.get_code(), Loader=yaml.FullLoader)


import os


MIKRO_J_PATH = os.path.dirname(os.path.realpath(__file__))

PLUGIN_PATH = os.getenv("MIKRO_J_PLUGINS_PATH", os.path.join(MIKRO_J_PATH, "plugins"))
MACROS_PATH = os.getenv("MIKRO_J_MACROS_PATH", os.path.join(MIKRO_J_PATH, "macros"))
ASSETS_PATH = os.path.join(MIKRO_J_PATH, "assets")


def get_asset_file(file, darkMode=False):
    return os.path.join(ASSETS_PATH, file)



import os
from typing import Tuple

from pydantic.main import BaseModel
from arkitekt.packers.structure import BoundType
from arkitekt.schema import Node
import re
from arkitekt.schema.ports import StructureArgPort, StructureReturnPort
from mikro.schema import Representation

from mikroj.registries.matcher import get_current_matcher

doc = re.compile("\/\*(?P<name>(.|\n)*)\*\/*")

documentation = re.compile(
    "\/\*\W*(?P<name>[^\n]*)\n\W*(?P<doc>[^\/\@]*)\n*\*.*"
)  # matches the first line in a docstring

in_re = re.compile("#@\W(?P<type>\w*)\W*(?P<key>\w*)")
return_re = re.compile("#@output\W(?P<type>\w*)\W*(?P<key>\w*)")

is_interactive_re = re.compile(".*@interactive*")
activein_re = re.compile(".*\@setactivein.*")
activeout_re = re.compile(".*\@takeactiveout*")
donecloseactive_re = re.compile(".*\@donecloseactive*")
filter_re = re.compile(".*\@filter*")
rgb_re = re.compile(".*\@rgb*")


params_re = re.compile(r"#@[^\(]*\((?P<params>[^\)]*)\)")  # line has params


class MacroDefinition(BaseModel):
    setactivein: bool = False  # mirorring CellProfiler approach
    takeactiveout: bool = False
    donecloseactive: bool = False
    interactive: bool = False
    filter: bool = False
    macro: str


def define_macro(file: str) -> Tuple[Node, MacroDefinition]:

    matcher = get_current_matcher()

    with open(file) as f:
        all = f.read()

    macro_def = MacroDefinition(macro=all)

    with open(file) as f:
        lines = f.readlines()

    x = documentation.match(all)

    macro_def.interactive = is_interactive_re.search(all) is not None
    macro_def.setactivein = activein_re.search(all) is not None
    macro_def.takeactiveout = activeout_re.search(all) is not None
    macro_def.donecloseactive = donecloseactive_re.search(all) is not None
    macro_def.filter = filter_re.search(all) is not None

    print("noainasoisnaoin", activein_re.search(all))
    if x:
        node_name = x.group("name")
        doc = x.group("doc").replace("*", "").replace("\n", "").strip()

    else:
        raise Exception("Not well defined Macro Documentation")

    args = []
    kwargs = []
    returns = []

    for line in lines:

        inmatch = in_re.match(line)
        if inmatch:
            imageJType = inmatch.group("type")
            key = inmatch.group("key")

            extras = {}
            description = ""

            paramsmatch = params_re.match(line)
            if paramsmatch:
                params = paramsmatch.group("params")
                for param in params.split(","):
                    pair = param.split("=")
                    extras[pair[0].strip()] = pair[1].strip()

            if "value" in extras:

                port = matcher.kwargbuilder_for_type(imageJType)(
                    key=key, description=description, default=extras["value"]
                )
                kwargs.append(port)

            else:
                port = matcher.argbuilder_for_type(imageJType)(
                    key=key, description=description
                )
                args.append(port)

        rematch = return_re.match(line)
        if rematch:
            imageJType = rematch.group("type")
            key = rematch.group("key")

            paramsmatch = params_re.match(line)
            if paramsmatch:
                params = paramsmatch.group("params")
                print(params)

            port = matcher.returnbuilder_for_type(imageJType)(
                key=key, description=description
            )
            returns.append(port)

            print("Return", type, key)

    if macro_def.setactivein:
        # Because this is how the fuckinng macros work apperently :rolling-eyes:
        args.insert(
            0,
            StructureArgPort(
                identifier="representation",
                key="image",
                widget=Representation.get_structure_meta().widget,
            ),
        )

    if macro_def.takeactiveout:
        returns.insert(
            0,
            StructureReturnPort(
                identifier="representation",
                key="image",
            ),
        )

    if macro_def.filter:
        assert (
            isinstance(args[0], StructureArgPort)
            and hasattr(args[0], "identifier")
            and args[0].identifier == "representation"
        ), "If Macro is a filter it should have as a first argument a Representation"

    return (
        Node(
            **{
                "name": node_name,
                "description": doc,
                "args": args,
                "kwargs": kwargs,
                "returns": returns,
                "package": "macros",
                "interface": os.path.basename(file),
            }
        ),
        macro_def,
    )




from mikroj.parsers.base import CodeBlockParser
from mikroj.parsers.code.base import CodeParser
from mikroj.registries.macro import register_code_parser


@register_code_parser("groovy")
class GroovyParser(CodeParser):
    language = "language-groovy"

    def __init__(self, *args, stream=None, **kwargs) -> None:
        super().__init__(*args, stream=stream, **kwargs)
        self.parser = CodeBlockParser(stream, self.language)

    def get_code(self):
        return self.parser.get_code()


from typing import Optional, Union

import pydantic
from arkitekt.schema.node import Node
from mikroj.actors.base import FuncMacroActor
from mikroj.actors.define_macro import MacroDefinition, define_macro
from mikroj.actors.default import DefaultMacroActor
from mikroj.parsers.base import Parser
from mikroj.parsers.definition.base import DefinitionParser
from mikroj.parsers.code.base import CodeParser
import pathlib
from functools import reduce
from mikroj.parsers.meta import MDMeta, parse_md_meta
from mikroj.registries.actor import get_current_macro_actor_registry
from pydantic import BaseModel
import logging
import os


logger = logging.getLogger(__name__)


class QueryNodeDefinition(BaseModel):
    package: Optional[str]
    interface: Optional[str]
    q: Optional[str]


class MacroRegistry:
    def __init__(self) -> None:
        self.registered_macros = {}
        self.registered_definition_parser = {}
        self.registered_code_parser = {}
        self.registered_definitions = {}

    def register_definition_parser(self, className, parser):
        self.registered_definition_parser[className] = parser

    def register_code_parser(self, className, parser):
        self.registered_code_parser[className] = parser

    def get_definition_for_interface(self, interface) -> MacroDefinition:
        return self.registered_definitions[interface]

    def scan_folder(self, folder_path="macros"):
        assert os.path.exists(folder_path), f"Folder {folder_path} does not exist"

        pathlist = pathlib.Path(folder_path).rglob("*.ijm")
        macro_list = []
        for path in pathlist:
            # because path is object not string
            path_in_str = str(path)
            logger.debug(f"Found file {path}")
            node, definition = define_macro(path_in_str)
            self.registered_definitions[node.interface] = definition
            actorBuilder = lambda: DefaultMacroActor()

            macro_list.append((node, actorBuilder))

        return macro_list


def register_definition_parser(className: str):
    def rea_decorator(parser):
        assert issubclass(parser, Parser), "Parser must subclass Parser"
        logger.info(f"Registering Definition Parser {parser} for {className}")
        get_current_macro_registry().register_definition_parser(className, parser)
        return parser

    return rea_decorator


def register_code_parser(className: str):
    def rea_decorator(parser):
        assert issubclass(parser, Parser), "Parser must subclass Parser"
        logger.info(f"Registering Code parser {parser} for {className}")
        get_current_macro_registry().register_code_parser(className, parser)
        return parser

    return rea_decorator


MACRO_REGISTRY = None


def get_current_macro_registry(register_defaults=True) -> MacroRegistry:
    global MACRO_REGISTRY
    if MACRO_REGISTRY is None:
        MACRO_REGISTRY = MacroRegistry()

    return MACRO_REGISTRY



from mikroj.parsers.base import Parser
from mikroj.parsers.definition.base import DefinitionParser
from mikroj.parsers.code.base import CodeParser
import pathlib
from functools import reduce
from mikroj.parsers.meta import parse_md_meta
import logging

logger = logging.getLogger(__name__)


class MacroActorRegistry:
    def __init__(self) -> None:
        self.registered_actors = {}

    def register_actor(self, className, parser):
        self.registered_actors[className] = parser

    def get_actor(self, className):
        return self.registered_actors[className]


def register_macro_actor(className: str):
    def real_decorator(macro_actor):
        logger.info(f"Registering Macro Actor {macro_actor} for {className}")
        get_current_macro_actor_registry().register_actor(className, macro_actor)
        return macro_actor

    return real_decorator


MACRO_ACTOR_REGISTRY = None


def get_current_macro_actor_registry(register_defaults=True) -> MacroActorRegistry:
    global MACRO_ACTOR_REGISTRY
    if MACRO_ACTOR_REGISTRY is None:
        MACRO_ACTOR_REGISTRY = MacroActorRegistry()
        if register_defaults:
            from mikroj.actors.default import StackToStackActor

    return MACRO_ACTOR_REGISTRY



class ImageMemory:
    def __init__(self) -> None:
        self.shelf = {}

    async def put(self, id, image):
        self.shelf[id] = image

    async def get(self, id):
        return self.shelf[id]


IMAGE_MEMORY = None


def get_current_memory():
    global IMAGE_MEMORY
    if not IMAGE_MEMORY:
        IMAGE_MEMORY = ImageMemory()
    return IMAGE_MEMORY



from mikroj.parsers.definition.base import DefinitionParser
import markdown

from mikroj.registries.macro import register_definition_parser


@register_definition_parser("meta")
class MetaParser(DefinitionParser):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.md = markdown.Markdown(extensions=["meta",'fenced_code'])
        self.html = self.md.convert(self.stream)
        self._definition = {}
        for key, item in self.md.Meta.items():
            if isinstance(item, list):
                item = item[0]
            self._definition[key] = item

    def get_definition(self):
        return self._definition



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mikroj',
 'mikroj.actors',
 'mikroj.parsers',
 'mikroj.parsers.code',
 'mikroj.parsers.definition',
 'mikroj.registries',
 'mikroj.structures',
 'mikroj.ui']

package_data = \
{'': ['*'], 'mikroj': ['assets/*', 'macros/*']}

install_requires = \
['Markdown>=3.3.4,<4.0.0',
 'PyQt5>=5.15.4,<6.0.0',
 'arkitekt>=0.1.106,<0.2.0',
 'mikro>=0.1.57,<0.2.0',
 'pyimagej>=1.0.1,<2.0.0']

entry_points = \
{'console_scripts': ['mikroj = mikroj.run:main']}

setup_kwargs = {
    'name': 'mikroj',
    'version': '0.1.18',
    'description': '',
    'long_description': None,
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from abc import abstractclassmethod, abstractmethod
from arkitekt.packers.structure import BoundType, Structure, StructureMeta
import uuid
from mikro.schema import Representation

from mikroj.structures.store import get_current_memory


class ImagePlus(Structure):
    """Image Plus BETA

    Fiji Image is an Agent bounc Structure that represent an in Memory open
    image on Fiji. It provides easy conversion to

    Args:
        Structure ([type]): [description]
    """

    def __init__(self, image, reference=None) -> None:
        self.image = image
        self.reference = reference or uuid.uuid4()
        super().__init__()

    @classmethod
    def get_structure_meta(cls) -> StructureMeta:
        return StructureMeta(bound=BoundType.AGENT, identifier="imageplus")

    @classmethod
    async def expand(cls, shrinked_value):
        return await get_current_memory().get(shrinked_value)

    async def shrink(self):
        return await get_current_memory().put(self.reference, self.image)



from threading import active_count
from typing import List
from arkitekt.schema.node import Node
from mikroj.actors.base import FuncMacroActor
from mikro import Representation
import dask
from mikroj.actors.define_macro import MacroDefinition

from mikroj.registries.helper import BaseImageJHelper, get_running_helper


def jtranspile(
    instance,
    helper: BaseImageJHelper,
    node: Node,
    macrodef: MacroDefinition,
    args: List,
    kwargs: dict,
):
    if isinstance(instance, Representation):
        x = helper.py.to_java(instance.data.squeeze().compute())
        print(x.__class__.__name__)
        return x

    return instance


def expand_xarray(xarray):
    if "Channel" in xarray.dims:
        xarray = xarray.rename({"Channel": "c"})
    if "z" not in xarray.dims:
        xarray = xarray.expand_dims("z")
    if "t" not in xarray.dims:
        xarray = xarray.expand_dims("t")
    if "c" not in xarray.dims:
        xarray = xarray.expand_dims("c")

    return xarray


def ptranspile(
    instance,
    helper: BaseImageJHelper,
    node: Node,
    macrodef: MacroDefinition,
    args: List,
    kwargs: dict,
):

    if instance.__class__.__name__ == "net.imagej.DefaultDataset":
        xarray = helper.py.from_java(instance)
        rep = Representation.objects.from_xarray(
            expand_xarray(xarray), name="Created from"
        )
        return rep

    if instance.__class__.__name__ == "ij.ImagePlus":
        xarray = helper.py.from_java(instance)

        if macrodef.filter:
            origin: Representation = args[0]
            rep = Representation.objects.from_xarray(
                expand_xarray(xarray),
                name=f"{node.name} of {origin.name}",
                origins=[origin],
                tags=["filtered", "mikroj"],
                meta={"mikroj:makro": node.interface},
                sample=origin.sample,
            )

        else:
            rep = Representation.objects.from_xarray(
                expand_xarray(xarray), name="Undetermined through MikroJ"
            )
        return rep

    return instance


class DefaultMacroActor(FuncMacroActor):
    def assign(self, *args, **kwargs):
        """[summary]

        Args:
            rep (Representation): [description]

        Returns:
            [type]: [description]


        """

        from mikroj.registries.macro import get_current_macro_registry

        definition = get_current_macro_registry().get_definition_for_interface(
            self.template.node.interface
        )

        helper = get_running_helper()

        imagej_args = {
            port.key: jtranspile(
                arg, helper, self.template.node, definition, args, kwargs
            )
            for arg, port in zip(args, self.template.node.args)
        }
        imagej_kwargs = {
            key: jtranspile(kwarg, helper, self.template.node, definition, args, kwargs)
            for key, kwarg in kwargs.items()
        }

        if definition.setactivein:
            print("Setting Active")
            image = imagej_args.pop("image")
            helper.ij.ui().show(args[0].name, image)

        print("Running Macro")
        value = helper.ij.py.run_macro(
            definition.macro, {**imagej_args, **imagej_kwargs}
        )
        print("Macro Finished")

        returns = []
        if definition.takeactiveout:
            print("Taking old Image out")
            returns.append(helper.py.active_image_plus())

        for index, re in enumerate(self.template.node.returns):
            if index == 0 and definition.takeactiveout:
                continue  # we already put the image out
            returns.append(value.getOutput(re.key))

        print("Transpiling Inputs")
        transpiled_returns = [
            ptranspile(value, helper, self.template.node, definition, args, kwargs)
            for value in returns
        ]

        if definition.donecloseactive:
            print("Closing")
            helper.ij.py.run_macro("close();", {})

        return tuple(transpiled_returns)



from arkitekt.schema import Node
from html.parser import HTMLParser
import markdown


class Parser:

    def __init__(self, *args, stream=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.stream = stream



class CodeBlockParser(HTMLParser):

    def __init__(self, stream, language, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.language = language
        self.note = False
        self.code_lines = []
        md = markdown.Markdown(extensions=["meta",'fenced_code'])
        html = md.convert(stream)
        self.feed(html)

    def handle_starttag(self, tag, attrs):
        if tag == "code" and self.language == dict(attrs)["class"]:
            self.note = True

    def handle_endtag(self, tag):
        if tag == "code": self.note = False

    def handle_data(self, data):
        if self.note:
            self.code_lines.append(data)


    def get_code(self):
        return "".join(self.code_lines)














from typing import Optional
import markdown
from pydantic import BaseModel

class MDMeta(BaseModel):
    title: str
    code_parser: Optional[str] = "groovy"
    definition_parser: Optional[str] = "yaml"
    actor: str

    def get_code_parsers(self):
        return[i.strip() for i in self.code_parser.split(",")]

    def get_definition_parsers(self):
        return [i.strip() for i in self.definition_parser.split(",")]




def parse_md_meta(stream):
    md = markdown.Markdown(extensions=["meta",'fenced_code'])
    html = md.convert(stream)
    _meta = {}
    for key, item in md.Meta.items():
        if isinstance(item, list):
            item = item[0]
        _meta[key] = item
    return MDMeta(**_meta)




from abc import abstractmethod
import abc
import sys
from PyQt5 import QtGui, QtWidgets
from PyQt5 import QtCore
from PyQt5.QtCore import QObject, QRect, Qt, pyqtSignal
import os
from pydantic.class_validators import extract_root_validators
from arkitekt.actors.registry import register
from arkitekt.qt.utils import register_ui
from herre.qt import QtHerre
from pydantic.main import BaseModel
from arkitekt.messages.base import T
from fakts.grants.qt.qtbeacon import QtSelectableBeaconGrant
from fakts.grants.qt.qtyamlgrant import QtYamlGrant
from fakts.grants.yaml import YamlGrant
from fakts.qt import QtFakts
from mikro import Representation
from mikroj.agent import MikroJAgent
from mikroj.env import MACROS_PATH, PLUGIN_PATH, get_asset_file
from mikroj.helper import ImageJHelper
from arkitekt.qt.agent import QtAgent
from mikro.widgets import MY_TOP_REPRESENTATIONS
from arkitekt.qt.widgets.provisions import ProvisionsWidget
from arkitekt.qt.widgets.templates import TemplatesWidget
from arkitekt.qt.widgets.magic_bar import MagicBar
import os

from mikroj.structures.imageplus import ImagePlus


packaged = False

if packaged:
    os.environ["JAVA_HOME"] = os.path.join(os.getcwd(), "share\\jdk8")
    os.environ["PATH"] = (
        os.path.join(os.getcwd(), "share\\mvn\\bin") + os.pathsep + os.environ["PATH"]
    )


helper = ImageJHelper()


class FormWrapped(QtWidgets.QWidget):
    def __init__(self, widget, title, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QHBoxLayout()
        self.formGroupBox = QtWidgets.QGroupBox(title)
        qlayout = QtWidgets.QFormLayout()
        qlayout.addRow(widget)
        self.formGroupBox.setLayout(qlayout)
        self.layout.addWidget(self.formGroupBox)
        self.setLayout(self.layout)


class ArkitektWidget(QtWidgets.QWidget):
    def __init__(self, helper, *args, config_path=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Different Grants

        self.beacon_grant = QtSelectableBeaconGrant()
        self.fakts = QtFakts(
            grants=[self.beacon_grant],
            subapp="mikroj",
            hard_fakts={
                "herre": {
                    "client_id": "hmtwKgUO092bYBOvHngL5HVikS2q5aWbS7V1ofdU",
                    "scopes": ["introspection", "can_provide"],
                }
            },
        )
        self.herre = QtHerre()
        self.agent = MikroJAgent(helper, self)

        self.magic_bar = MagicBar(self.fakts, self.herre, self.agent)
        self.agent.load_macros(MACROS_PATH)

        self.layout = QtWidgets.QVBoxLayout()

        self.provisions_widget = FormWrapped(ProvisionsWidget(self.agent), "Provisions")
        self.templates_widget = FormWrapped(TemplatesWidget(self.agent), "Templates")

        self.layout.addWidget(self.magic_bar)
        self.layout.addWidget(self.provisions_widget)
        self.layout.addWidget(self.templates_widget)
        self.setLayout(self.layout)


def show_image(rep: Representation):
    """Shows an Image

    Shows a Representation on Imagej

    Args:
        rep (Representation): A Beautiful Little Image to display
    """


@register(interfaces=["bridge"])
def bridge(rep: Representation) -> ImagePlus:
    """Bridges a thing to a thang

    Bridges an Representation to an In memory Fiji Image

    Args:
        rep (Representation): [description]

    Returns:
        ImagePlus: [description]
    """
    print(rep)
    array = rep.data.compute()
    print(array)

    plus = ImagePlus(array)
    print(plus)
    return plus


class MikroJ(QtWidgets.QMainWindow):
    def __init__(self, **kwargs):
        super().__init__()
        # self.setWindowIcon(QtGui.QIcon(os.path.join(os.getcwd(), 'share\\assets\\icon.png')))
        self.setWindowIcon(QtGui.QIcon(get_asset_file("logo.ico")))

        self.arkitektWidget = ArkitektWidget(helper, **kwargs)

        self.agent = self.arkitektWidget.agent

        self.showActor = self.agent.register_ui(
            show_image, widgets={"rep": MY_TOP_REPRESENTATIONS}
        )

        self.showActor.signals.assign.wire(self.show_image_assign)

        self.layout = QtWidgets.QHBoxLayout()
        self.setCentralWidget(self.arkitektWidget)
        self.init_ui()

    def show_image_assign(self, res, args, kwargs):
        helper.displayRep(args[0])
        self.showActor.signals.assign.resolve(res, None)

    def init_ui(self):
        self.setWindowTitle("MikroJ")
        self.show()


def main(**kwargs):
    app = QtWidgets.QApplication(sys.argv)
    # app.setWindowIcon(QtGui.QIcon(os.path.join(os.getcwd(), 'share\\assets\\icon.png')))
    main_window = MikroJ(**kwargs)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



from arkitekt.qt.agent import QtAgent
from arkitekt.schema.node import Node
from mikroj.registries.macro import get_current_macro_registry, QueryNodeDefinition
import logging

logger = logging.getLogger(__name__)


class MikroJAgent(QtAgent):
    def __init__(self, helper, *args, strict=False, **kwargs) -> None:
        super().__init__(*args, strict=strict, **kwargs)
        self.helper = helper

    def load_macros(self, filepath):
        logger.debug(f"Opening Macros in {filepath}")
        macro_associations = get_current_macro_registry().scan_folder(
            folder_path=filepath
        )

        for defi, actorBuilder in macro_associations:
            if isinstance(defi, QueryNodeDefinition):
                self.registry.templatedUnqueriedNodes.append(
                    (defi.dict(), actorBuilder, {})
                )
            if isinstance(defi, Node):
                self.registry.templatedNewNodes.append((defi, actorBuilder, {}))





from mikroj.parsers.base import Parser
from html.parser import HTMLParser
import markdown

class CodeParser(Parser):

    def __init__(self, *args, stream=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.stream = stream

    def get_code(self):
        raise NotImplementedError()



